/*
 * fanxdell
 * Userspace fancontrol for use while installing Debian.
 * Copyright © 2014 C_Xperimental
 * Not associated in any way with the company Dell.
 * Not associated in any way with i8kutils or dellfand, although several
 * functions have been adapted from their source code.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * <www.cxperimental.weebly.com>
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>
#include <argp.h>

/* Uncomment line below for debug, try tail -f /var/log/syslog for daemon. */
/*#define DEBUG 1*/

/* Uncomment line below to disable make/model safety checks, then compile. */
/*#define UNSAFE_MODE 1*/

/* Uncomment to enable non root use of --info with 's' bit set */
/*#define SET_EUID 1*/

#define PID_FILE_NAME "fanxdell_8sl7m21wdym4.pid"
#define DMI_MAKE "/sys/class/dmi/id/sys_vendor"
#define DMI_MODEL "/sys/class/dmi/id/product_name"
#define PROGRAM_VERSION 1

/* Signal interrupt handling */
#define X_FINITE_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int result;                                                       \
       long int index = 0;                                                    \
       do result = (long int) (expression);                                   \
       while (result == -1L && errno == EINTR && (index++) < 1000);          \
       result; }))

typedef struct
{
  unsigned int eax;
  unsigned int ebx __attribute__ ((packed));
  unsigned int ecx __attribute__ ((packed));
  unsigned int edx __attribute__ ((packed));
  unsigned int esi __attribute__ ((packed));
  unsigned int edi __attribute__ ((packed));
} SMMRegisters;

struct fan_data
{
  int what_fan; /* 0 left, 1 right, 2 both */
  int poll;
  int t01;
  int t02;
  int h01;
  int h02;
  int t11;
  int t12;
  int h11;
  int h12;
};

void sigf (char *sig, int err);
void signal_setup (void);
void signal_unblock (void);
void kill_option (void) __attribute__ ((noreturn));
int lock_n_kill (void);
int dell_smm (SMMRegisters *regs);
int get_cpu_temp (void);
int get_fan_status (int fan);
int set_fan (int fan, int speed);
int get_bios_version (void);
int get_fan_speed (int fan);
int get_dell_sig (int cmd, int *eax, int *edx);
int get_docking_state (void);
#ifndef UNSAFE_MODE
void dmi_check (void);
#endif
void get_pid_dir (void);
void set_lock (int how);
void pid_file (void);
void fan_daemon (void) __attribute__ ((noreturn, hot));
void typo (void) __attribute__ ((noreturn));
int main (int argc, char **argv);

/* Global variables for signal handling */
volatile sig_atomic_t usr1 = 0, usr2 = 0, fin = 0, newpid = 0, sigsave = 0;

/* More global variables */
char *pid_dir;
int fd;
struct fan_data dat =
{
  .what_fan = 0,
  .poll = 10,
  .t01 = 35, /* fan0 if using 2 fans, 'the' fan if using only one */
  .t02 = 62, /* t = temperature threshold */
  .h01 = 5,  /* h = hysteresis buffer */
  .h02 = 8,
  .t11 = 47, /* fan1 if using 2 fans */
  .t12 = 62,
  .h11 = 7,
  .h12 = 8
};
char usage[] =
"fanxdell --help / -?  : prints this help.\n\n\
fanxdell --fan / -f  [-/0/1/2] [-/0/1/2]  : alters fan speed\n\n\
fanxdell --info / -i  : prints system info to stdout\n\n\
fanxdell --daemon / -d fan poll t01 t02 h01 h02 t11 t12 h11 h12\n\
Starts an automatic fan control program, or replaces a running fanxdell.\n\
fan is the only required argument. Select 0, 1, or for twin fans use 2.\n\
t01 = turn on fan 0 at low speed at this temperature.\n\
h12 = hysteresis for high speed threshold (t12) for fan 1, and so on.\n\
Any preceding arguments must be given to set later values.\n\
Default settings for single fan:\n\
fanxdell --daemon [0/1] 10 35 62 5 8\n\
Default settings for twin fan:\n\
fanxdell --daemon 2 10 35 62 5 8 47 62 7 8\n\
Unset values stay at defaults.\n\
Temperature thresholds and hysteresis values must be between 0 and 127 c.\n\
Polling frequency must be between 1 and 30 seconds.\n\
Simplest use: fanxdell -d [0/1/2]\n\n\
fanxdell --auto / -a poll t01 t02 h01 h02 t11 t12 h11 h12\n\
Identical to --daemon option but automatically detects fans.\n\
Simplest use:  fanxdell -a\n\n\
fanxdell --kill / -k  : Stops any running fanxdell daemon.\n\n\
fanxdell --version / -V  : prints program version.\n\
\n";

/* Signal handler functions */
void
usr1_handler (int signum, siginfo_t *info, void *ptr __attribute__ ((unused)))
{
  usr1++;
  sigsave = signum;
  newpid = info->si_pid;
}

void
usr2_handler (int signum __attribute__ ((unused)))
{
  usr2++;
}

void
fin_handler (int signum)
{
  fin++;
  sigsave = signum;
}

void
sigf (char *sig, int err)
{
  errno = err;
  fprintf (stderr, "%s, sigaction failure : %m\n", sig);
  exit (EXIT_FAILURE);
}

/* Setup signal handler functions for various signals */
void
signal_setup (void)
{
  struct sigaction newaction, oldaction;
  newaction.sa_sigaction = usr1_handler;
  sigemptyset (&newaction.sa_mask);
  newaction.sa_flags = SA_RESTART | SA_SIGINFO; /* SA_SIGINFO for 3 args */
  errno = 0;
  if (sigaction (SIGUSR1, &newaction, NULL)) sigf ("SIGUSR1", errno);
  newaction.sa_handler = usr2_handler;
  newaction.sa_flags = SA_RESTART;
  errno = 0;
  if (sigaction (SIGUSR2, &newaction, NULL)) sigf ("SIGUSR2", errno);
  newaction.sa_handler = fin_handler;
  oldaction.sa_handler = NULL;
  sigemptyset (&oldaction.sa_mask);
  errno = 0;
  if (sigaction (SIGTERM, NULL, &oldaction)) sigf ("SIGTERM query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGTERM, &newaction, NULL)) sigf ("SIGTERM", errno);
    }
  oldaction.sa_handler = NULL;
  errno = 0;
  if (sigaction (SIGINT, NULL, &oldaction)) sigf ("SIGINT query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGINT, &newaction, NULL)) sigf ("SIGINT", errno);
    }
  oldaction.sa_handler = NULL;
  errno = 0;
  if (sigaction (SIGQUIT, NULL, &oldaction)) sigf ("SIGQUIT query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGQUIT, &newaction, NULL)) sigf ("SIGQUIT", errno);
    }
}

/* Helper function used during cleanup when daemon exits */
void
signal_unblock (void)
{
 struct sigaction newaction;
  newaction.sa_handler = SIG_DFL;
  sigemptyset (&newaction.sa_mask);
  errno = 0;
  if (sigaction (sigsave, &newaction, NULL))
    {
      syslog (LOG_ERR, "sigaction failure at cleanup : %m");
      exit (EXIT_FAILURE);
    }
}

/* Manages --kill option */
void
kill_option (void) /* Find pidfile(s) first */
{
  struct stat test_stat;
  /* Allocate memory: pid name length + directory (11) + null (1) */
  /* If the directory names are changed this may need changing. */
  signal_setup ();
  errno = 0;
  pid_dir = (char *) malloc ((size_t) ((strlen (PID_FILE_NAME)) + 12));
  if (pid_dir == NULL)
    {
      perror ("Virtual memory exhausted in kill_option");
      exit (EXIT_FAILURE);
    }
  if ((!(stat ("/run/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode))) /* Is it a directory? */
    {
      strcpy (pid_dir, "/run/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* Its a pidfile */
          if (lock_n_kill ())
            fprintf (stderr, "Error processing %s\n", pid_dir);
        }
    }
  if ((!(stat ("/tmp/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode)))
    {
      strcpy (pid_dir, "/tmp/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* Its a pidfile */
          if (lock_n_kill ())
            fprintf (stderr, "Error processing %s\n", pid_dir);
        }
    }


  strcpy (pid_dir, "/");
  strcat (pid_dir, PID_FILE_NAME);
  if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
    {
      /* Its a pidfile */
      if (lock_n_kill ())
        fprintf (stderr, "Error processing %s\n", pid_dir);
    }
  exit (EXIT_SUCCESS);
}

/* Helper function for kill_option */
int
lock_n_kill (void)
{
  int i;
  struct flock lk_dat;
  /* Open and get a file descriptor */
  fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDWR | O_NONBLOCK,
                               S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      perror ("Pidfile open failed");
      fprintf (stderr, "Could not get a file descriptor for %s\n", pid_dir);
      return (1); /* Because alternate directories can be checked */
    }
  for (i = 3; i >= 1; i--)
    {
      lk_dat.l_type = F_WRLCK;
      lk_dat.l_whence = SEEK_SET;
      lk_dat.l_start = (off_t) 0;
      lk_dat.l_len = (off_t) 1;
      lk_dat.l_pid = (pid_t) 0;
      errno = 0;
      if (fcntl (fd, F_GETLK, (&lk_dat)))
        {
          perror ("fcntl F_GETLK failed");
          fprintf (stderr, "Could not get lock status for %s\n", pid_dir);
          return (1);
        }
      if (lk_dat.l_type == F_UNLCK)
        {
          /* No active lock, stale pidfile */
          if (i == 3) printf ("Unlocked pidfile, checking for daemon -\n");
          printf ("Seconds remaining : %d\n", i);
          /* Loop in case of a daemon being started during this check */
          if (i <= 1)
            {
              errno = 0;
              if (unlink (pid_dir))
                {
                  perror ("First unlink failed in lock_n_kill");
                  fprintf (stderr, "Couldn't delete %s\n", pid_dir);
                  return (1);
                }
              printf ("Deleted unlocked pidfile : %s\n", pid_dir);
              return (0);
             }
        }
      if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
        {
          break;
        }
      sleep (1);
    } /* end of for loop */

  if (lk_dat.l_type != F_WRLCK)
    {
      fprintf (stderr, "File %s has an unexpected lock status\n", pid_dir);
      return (1);
    }
  printf ("Active lock on %s\n", pid_dir);
  /* File locked, kill daemon */
  errno = 0;
  if (kill (lk_dat.l_pid, SIGUSR1))
    {
      perror ("kill failed to send signal");
      return (1);
    }
  for (i = 5; i >= 1; i--)
    {
      if (usr2 > 0)
        {
          /* Previous daemon has released lock, delete pidfile */
          printf ("Daemon returned SIGUSR2, unlinking %s\n", pid_dir);
          if (unlink (pid_dir))
            {
              perror ("Second unlink failed in lock_n_kill");
              fprintf (stderr, "Couldn't delete %s\n", pid_dir);
              return (1);
            }
          return (0);
        }
      if (i == 5) printf ("Waiting for SIGUSR2\n");
      printf ("Seconds to timeout = %d\n", i);
      sleep (1);
      /* Otherwise keep looping until SIGUSR2 is received */
    }
  /* Unknown proc locks pidfile, fanxdell malfunction, CPU o/load? */
  printf ("%s locked by unresponsive process\npid no. = %d\n",
          pid_dir, lk_dat.l_pid);
  if (kill (lk_dat.l_pid, SIGKILL))
    {
      perror ("kill failed to send SIGKILL");
      return (1);
    }
  printf ("SIGKILL sent to %d\n", lk_dat.l_pid);
  printf ("Unlinking %s\n", pid_dir);
  if (unlink (pid_dir))
    {
      perror ("Third unlink failed in lock_n_kill");
      fprintf (stderr, "Couldn't delete %s\n", pid_dir);
      return (1);
    }
  return (0);
}

/* Write 1 byte to 0xb2, 1 byte to 0x84, put CPU in system management mode */
/* 0xb2 = Dell command register, 0x84 = ACPI SMB_DATA[0] Data register zero */
int
dell_smm (SMMRegisters *regs)
{
  int rc;
  unsigned int eax = regs->eax;
  asm volatile ("pushl %%eax\n\t" \
      "movl 0(%%eax),%%edx\n\t" \
      "push %%edx\n\t" \
      "movl 4(%%eax),%%ebx\n\t" \
      "movl 8(%%eax),%%ecx\n\t" \
      "movl 12(%%eax),%%edx\n\t" \
      "movl 16(%%eax),%%esi\n\t" \
      "movl 20(%%eax),%%edi\n\t" \
      "popl %%eax\n\t" \
      "out %%al,$0xb2\n\t" \
      "out %%al,$0x84\n\t" \
      "xchgl %%eax,(%%esp)\n\t"
      "movl %%ebx,4(%%eax)\n\t" \
      "movl %%ecx,8(%%eax)\n\t" \
      "movl %%edx,12(%%eax)\n\t" \
      "movl %%esi,16(%%eax)\n\t" \
      "movl %%edi,20(%%eax)\n\t" \
      "popl %%edx\n\t" \
      "movl %%edx,0(%%eax)\n\t" \
      "lahf\n\t" \
      "shrl $8,%%eax\n\t" \
      "andl $1,%%eax\n" \
      : "=a" (rc)
      : "a" (regs)
      : "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory");
  if ((regs->eax & 0xffff) == 0xffff) return (-EINVAL);
  /* The above case MUST be handled separately for auto fan detection */
  if ((rc != 0) || (regs->eax == eax)) return (-EAGAIN);
  return (0);
}  /* i386: EINVAL = 22 EAGAIN = 11 (- = no confusion with valid result) */

/* Read the cpu temperature. */
int
get_cpu_temp (void)
{
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  int temp, err;
  /* Set to 0 on initialisation only. */
  static int prev = 0;
  regs.eax = 0x10a3;
  err = dell_smm (&regs);
  if (err) return (err);
  temp = regs.eax & 0xff;
  /* Sometimes the temperature sensor returns 0x99, which is out of range. */
  /* In this case we return (once) the previous cached value. */
  if (temp > 127)
    {
      temp = prev;
      prev = 127;
    }
  else
    {
      prev = temp;
    }
  return (temp);
}

/* Current fan setting, off, low, or high: ie. 0, 1, or 2 */
int
get_fan_status (int fan)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  regs.eax = 0x00a3;
  regs.ebx = fan & 0xff;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax & 0xff);
}

/* Alters fan speed, fan = 0 or 1, speed = 0, 1 or 2 */
int
set_fan (int fan, int speed)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  /* set the register values with command and argument */
  regs.eax = 0x01a3;
  regs.ebx = (fan & 0xff) | (speed << 8);
  /* call the assembly code function */
  err = dell_smm (&regs);
  return (err);
}

/* Read the BIOS version */
int
get_bios_version (void)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x00a6;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax);
}

/* Read the fan speed in RPM */
int
get_fan_speed (int fan)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x02a3;
  regs.ebx = fan & 0xff;
  err = dell_smm (&regs);
  if (err) return (err);
  return ((regs.eax & 0xffff) * 15); /* 15 = fan speed multiplier */
}

/* Read DIAG DELL signatures from BIOS */
int
get_dell_sig (int cmd, int *eax, int *edx)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = cmd;
  err = dell_smm (&regs);
  if (err) return (err);
  *eax = regs.eax;
  *edx = regs.edx;
  return (0);
}

/* Check if laptop is docked */
int
get_docking_state (void)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x40a3;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax & 0xffff);
}

/* Read sysfs kernel files to check this is a compatible machine */
#ifndef UNSAFE_MODE
void
dmi_check (void)
{
  char *str1 = NULL;
  ssize_t get_ret = 0;
  size_t n_bytes = (size_t) 0;
  FILE *stream1;
  int dell = 0;
  int livp = 0;

  errno = 0;
  stream1 = fopen (DMI_MAKE, "r");
  if (stream1 == NULL)
    {
      fprintf (stderr, "Could not open %s\n", DMI_MAKE);
      perror ("fopen");
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check make is Dell, strstr returns null for no match */
      if (strstr (str1, "Dell"))
        {
#ifdef DEBUG
          puts ("Well, it's a Dell");
#endif
          dell = 1;
          break;
        }
    }
  while (get_ret != -1);
  /* Runs through loop twice, not changing str1 2nd time */
  free (str1);
  str1 = NULL;
  fclose (stream1);
  n_bytes = 0;
  errno = 0;
  stream1 = fopen (DMI_MODEL, "r");
  if (stream1 == NULL)
    {
      fprintf (stderr, "Could not open %s\n", DMI_MODEL);
      perror ("fopen");
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check for supported model name */
      if ((strstr (str1, "Latitude")) || (strstr (str1, "Inspiron"))
          || (strstr (str1, "Precision")) || (strstr (str1, "MM061"))
          || (strstr (str1, "MP061")) || (strstr (str1, "Vostro")))
        {
#ifdef DEBUG
          puts ("It's a Latitude, Inspiron, Vostro or Precision");
#endif
          livp = 1;
          break;
        }
    }
  while (get_ret != -1);
  if (dell == 0)
    {
      fprintf (stderr, "Cannot find Dell in %s\n", DMI_MAKE);
    }
  if (livp == 0)
    {
      fprintf (stderr,
               "Latitude, Inspiron, Precision, Vostro, MM061, MP061\
                not found in %s\n", DMI_MODEL);
    }
  if ((dell == 0) || (livp == 0))
    {
      fprintf (stderr,
               "Your computer may be incompatible, try further research\n");
      exit (EXIT_FAILURE);
    }
}
#endif

/* Location for pidfile - /run/local, /tmp/local, or / as a last resort */
void
get_pid_dir (void)
{
  struct stat test_stat;
  /* Allocate memory: pid name length + directory (11) + null (1) */
  /* If the directories are changed then this may need changing too. */
  pid_dir = (char *) malloc ((strlen (PID_FILE_NAME)) + 12);
  if (pid_dir == 0)
    {
      syslog (LOG_ERR, "malloc: Virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  if ((stat ("/run", &test_stat)) == 0)
    {
      if ((!(stat ("/run/local", &test_stat))) &&
          (S_ISDIR (test_stat.st_mode))) /* is it a directory? */
        {
          strcpy (pid_dir, "/run/local/");
          return;
        }
      if (!(mkdir ("/run/local", S_IRWXU | S_IRGRP | S_IXGRP
          | S_IROTH | S_IXOTH)))
          /* fails if /run wasn't a dir */
        {
          strcpy (pid_dir, "/run/local/");
          return;
        }
      syslog (LOG_ERR, "mkdir: failed to create /run/local");
    }
  if ((stat ("/tmp", &test_stat)) == 0)
    {
      if ((!(stat ("/tmp/local", &test_stat))) &&
          (S_ISDIR (test_stat.st_mode)))
        {
          strcpy (pid_dir, "/tmp/local/");
          return;
        }
      if (!(mkdir ("/tmp/local", S_IRWXU | S_IRGRP | S_IXGRP
          | S_IROTH | S_IXOTH)))
        {
          strcpy (pid_dir, "/tmp/local/");
          return;
        }
      syslog (LOG_ERR, "mkdir: failed to create /tmp/local");
    }
  if (mkdir ("/run", S_IRWXU | S_IRGRP | S_IXGRP
      | S_IROTH | S_IXOTH))
    {
      syslog (LOG_ERR, "mkdir: failed to create /run");
      strcpy (pid_dir, "/");
      return;
    }
  if (mkdir ("/run/local", S_IRWXU | S_IRGRP | S_IXGRP
      | S_IROTH | S_IXOTH))
    {
      syslog (LOG_ERR, "mkdir: failed to create /run/local");
      strcpy (pid_dir, "/");
      return;
    }
  strcpy (pid_dir, "/run/local/");
  return;
}

/* how = 1 for write lock, how = 2 to unlock */
void
set_lock (int how)
{
  struct flock lk_dat;
  /* fcntl with these cmds returns 0, or -1 with errno set. */
  if (how == 1) lk_dat.l_type = F_WRLCK;
  if (how == 2) lk_dat.l_type = F_UNLCK;
  lk_dat.l_whence = SEEK_SET;
  lk_dat.l_start = (off_t) 0;
  lk_dat.l_len = (off_t) 1;
  errno = 0;
  if (fcntl (fd, F_SETLK, (&lk_dat)))
    {
      syslog (LOG_ERR, "fcntl F_SETLK failed. Error = %m");
      if (how == 1) exit (EXIT_FAILURE);
      /* For unlock, the fd will get closed next anyway. */
    }
}

/* Create and lock pidfile or get pid of existing fan daemon */
void
pid_file (void)
{
  int count = 0;
  struct flock lk_dat;
  get_pid_dir ();
  strcat (pid_dir, PID_FILE_NAME);
  errno = 0;
  fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDWR | O_CREAT | O_EXCL |
                                O_NONBLOCK, S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      /* file not opened: an error, or it exists */
      if ((errno == EEXIST) || (errno == EWOULDBLOCK))
        {
          /* probably pidfile already exists */
          errno = 0;
          /* Last open failed, so get a valid file descriptor */
          fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDWR | O_NONBLOCK,
                                        S_IWUSR | S_IRUSR));
          if (fd < 0)
            {
              syslog (LOG_ERR, "Second pidfile open failed. Error = %m");
              exit (EXIT_FAILURE);
            }
          lk_dat.l_type = F_WRLCK;
          lk_dat.l_whence = SEEK_SET;
          lk_dat.l_start = (off_t) 0;
          lk_dat.l_len = (off_t) 1;
          lk_dat.l_pid = (pid_t) 0;
          errno = 0;
          if (fcntl (fd, F_GETLK, (&lk_dat)))
            {
              syslog (LOG_ERR, "fcntl F_GETLK failed. Error = %m");
              exit (EXIT_FAILURE);
            }
          if (lk_dat.l_type == F_UNLCK)
            {
              /* No active lock, stale pidfile, take over and lock it */
#ifdef DEBUG
              syslog (LOG_ERR, "Stale pidfile found, fd = %d", fd);
#endif
              set_lock (1);
              return;
            }
          if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
            {
              /* File locked, replace previous daemon */
              errno = 0;
              if (kill (lk_dat.l_pid, SIGUSR1))
                {
                  syslog (LOG_ERR, "kill failed in pid_file. Error = %m");
                  exit (EXIT_FAILURE);
                }
              while (count < 60)
                {
                  count++;
                  sleep (1);
                  if (usr2 > 0)
                    {
                      /* Previous daemon has released pidfile lock */
                      set_lock (1);
#ifdef DEBUG
                      syslog (LOG_ERR, "Daemon returned SIGUSR2, fd = %d", fd);
#endif
                      return;
                    }
                  /* Otherwise keep looping until SIGUSR2 is received */
                }
              /* Multiple daemons started, unknown process locked pidfile? */
              syslog (LOG_ERR, "Pidfile locked, no response, timed out.");
              exit (EXIT_FAILURE);
            }
          /* Not unlocked, not F_WRLCK, something has a F_RDLCK? */
          syslog (LOG_ERR, "Failure: pidfile locked by another process?");
          exit (EXIT_FAILURE);
        }
      /* errno says it is a definite error */
      syslog (LOG_ERR, "open failed in pid_file. Error = %m");
      exit (EXIT_FAILURE);
    }
  /* open succeeded, set lock to prove this process is active */
  set_lock (1);
#ifdef DEBUG
  syslog (LOG_ERR, "No pidfile, new start, fd = %d", fd);
#endif
  return;
}

/* --daemon option */
void
fan_daemon (void)
{
  pid_t pid, sid;
  int tmp = 0, cpu_t = 0, f_stat0 = 0, f_stat1 = 0;

#ifdef DEBUG
  printf ("Preparing to fork, pid file name = %s\n", PID_FILE_NAME);
#endif
  signal_setup ();
  /* fork and write a pid file to /(run or tmp) /local/ or / */
  pid = fork ();
  if (pid < 0)
    {
      perror ("fork in fan_daemon failed");
      exit (EXIT_FAILURE);
    }
  if (pid > 0) exit (EXIT_SUCCESS);
  umask (0);
  /* open syslog and set values, /dev/log is opened by syslog if needed */
  openlog ("fanxdell", LOG_PID, LOG_DAEMON);
  sid = setsid();
  if (sid < 0)
    {
      syslog (LOG_ERR, "setsid failed. Error =%m");
      exit (EXIT_FAILURE);
    }
  if ((chdir ("/")) < 0)
    {
      syslog (LOG_ERR, "chdir failed. Error =%m");
      exit (EXIT_FAILURE);
    }
  fclose (stdin);
  fclose (stdout);
  fclose (stderr);
  /* The pidfile's FD will be 1 so don't write to stdout now! */
  /* The file lock would not survive the fork, so now: */
  pid_file ();
  /* Ensure temp2 is the higher value */
  if (dat.t01 > dat.t02)
    {
      tmp = dat.t01;
      dat.t01 = dat.t02;
      dat.t02 = tmp;
    }
  if (dat.t11 > dat.t12)
    {
      tmp = dat.t11;
      dat.t11 = dat.t12;
      dat.t12 = tmp;
    }
  /* Create a 5 C gap between the temperature thresholds if needed */
  if ((dat.t02 - dat.t01) < 5) dat.t01 = (dat.t02 - 5);
  if ((dat.t12 - dat.t11) < 5) dat.t11 = (dat.t12 - 5);
#ifdef DEBUG
  syslog (LOG_ERR, "About to start daemon loop");
#endif
  /* Start the daemon loop */
  while (1)
    {
      tmp = cpu_t; /* Save old temperature in case of strange errors */
      cpu_t = get_cpu_temp ();
      if ((cpu_t == -EINVAL) || (cpu_t == -EAGAIN)) cpu_t = tmp;
#ifdef DEBUG
        syslog (LOG_ERR, "t01=%d, t02=%d, h01=%d, h02=%d, p=%d, cpu=%d",
                dat.t01, dat.t02, dat.h01, dat.h01, dat.poll, cpu_t);
        syslog (LOG_ERR, "t11=%d, t12=%d, h11=%d, h12=%d, what_fan=%d",
                dat.t11, dat.t12, dat.h11, dat.h12, dat.what_fan);
#endif
      /* Check the current fan speed setting */
      if (dat.what_fan < 2) f_stat0 = get_fan_status (dat.what_fan);
      if (dat.what_fan == 2) /* Determine fan 1 setting for 2 fans */
        {
          f_stat0 = get_fan_status (0); /* Error only loses hysteresis */
          f_stat1 = get_fan_status (1);
          tmp = f_stat1; /* As below for f_stat0 */
          if (((cpu_t < dat.t11) && (f_stat1 == 0)) ||
              (cpu_t < (dat.t11 - dat.h11)))
            f_stat1 = 0;
          else f_stat1 = 1;
          if (((cpu_t >= (dat.t12 - dat.h12)) && (tmp == 2)) ||
              (cpu_t > dat.t12))
            f_stat1 = 2;
        } /* Now find a setting for fan 0 (0 or 1 for a single fan) */
      tmp = f_stat0; /* tmp needed for old f_stat0 value */
      /* What should the fan(s) be set to? */
      if (((cpu_t < dat.t01) && (f_stat0 == 0)) ||
          (cpu_t < (dat.t01 - dat.h01)))
        f_stat0 = 0;
      else f_stat0 = 1; /* If it isn't 0 then it's 1 */
      if (((cpu_t >= (dat.t02 - dat.h02)) && (tmp == 2)) ||
          (cpu_t > dat.t02))
        f_stat0 = 2; /* But 1 may get changed to 2 here */
      /* Set the fan(s) */
      if (dat.what_fan < 2)
        {
#ifdef DEBUG
          syslog (LOG_ERR, "Setting fan %d to %d", dat.what_fan, f_stat0);
          if (set_fan (dat.what_fan, f_stat0))
            syslog (LOG_ERR, "Error from set_fan");
#else
          set_fan (dat.what_fan, f_stat0);
#endif
        }
      else
        {
#ifdef DEBUG
          syslog (LOG_ERR, "Setting fan0: %d, fan1: %d", f_stat0, f_stat1);
          if (set_fan (0, f_stat0))
              syslog (LOG_ERR, "Error from set_fan 0");
          if (set_fan (1, f_stat1))
              syslog (LOG_ERR, "Error from set_fan 1");
#else
          set_fan (0, f_stat0); /* An error would mean the fan wasn't set */
          set_fan (1, f_stat1); /* But it'll get tried again next time round */
#endif
        }
      sleep (dat.poll);
      /* 'sleep' will be interrupted by any signals arriving */
      if ((usr1 > 0) || (fin > 0)) break;
    }
  /* Daemon loop ended, clean up */
  /* If a new fanxdell daemon opens syslog the old daemon's messages are */
  /* then not shown in Debian 7, using rsyslog. */
  set_lock (2);
  errno = 0;
  if (X_FINITE_FAILURE_RETRY (close (fd)))
    {
      syslog (LOG_ERR, "close failed in cleanup, error = %m");
    }
  /* When a new daemon has a descriptor, unlink makes a ghost file, so: */
  if (usr1 == 0)
    {
      errno = 0;
      if (unlink (pid_dir))
        {
          syslog (LOG_ERR, "unlink failed in cleanup, error = %m");
        }
    }
  /* Signal replacement daemon to proceed */
  if (usr1 > 0)
    {
      errno = 0;
      if (kill (newpid, SIGUSR2))
        {
          syslog (LOG_ERR, "kill failed to send SIGUSR2, error = %m");
        }
    }
  /* Unblock then raise signal for correct exit code */
  signal_unblock ();
  raise (sigsave);
  syslog (LOG_ERR, "Fatal signal not fatal - terminating.");
  exit (EXIT_FAILURE);
}

void
typo (void)
{
  fprintf(stderr, usage);
  exit (EXIT_FAILURE);
}

/* Main function, program execution begins here */
int
main(int argc, char **argv)
{
  int val1 = 0, val2 = 0, tmp1 = 0, tmp2 = 0, err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
#ifdef DEBUG
  puts ("Compiled for debug mode, writing to stdout and / or syslog");
#endif
  /* --help option (argp doesn't work in installation environment) */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--help")))
      || (!(strcmp (argv[1], "-?")))))
    {
      fprintf(stderr, usage);
      exit (EXIT_SUCCESS);
    }
  /* --version option */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--version")))
      || (!(strcmp (argv[1], "-V")))))
    {
      printf ("fanxdell version number = %d\n", PROGRAM_VERSION);
      exit (EXIT_SUCCESS);
    }
#ifndef UNSAFE_MODE
  dmi_check ();
#endif
#ifdef SET_EUID
  /* Non root use of -i with file owner = root, chmod u+s .../fanxdell */
  if (!((argc >= 2) && ((!(strcmp (argv[1], "--info")))
      || (!(strcmp (argv[1], "-i"))))))
    {
       errno = 0;
       if (seteuid (getuid()))
         {
           perror ("seteuid failure in main");
         }
    }
  /* Root permission still needed for all fan setting options */
#endif
  /* Ensure non root incorrect options get correct message with lengthy if */
  if (((argc >= 2) && ((!(strcmp (argv[1], "--info")))
      || (!(strcmp (argv[1], "-i")))))
      || ((argc >= 3) && ((!(strcmp (argv[1], "--daemon")))
      || (!(strcmp (argv[1], "-d")))))
      || ((argc >=2) && ((!(strcmp (argv[1], "--auto")))
      || (!(strcmp (argv[1], "-a")))))
      || ((argc >= 3) && ((!(strcmp (argv[1], "--fan")))
      || (!(strcmp (argv[1], "-f")))))
      || ((argc >= 2) && ((!(strcmp (argv[1], "--kill")))
      || (!(strcmp (argv[1], "-k"))))))
    {
      errno = 0;
      if (ioperm (0x84, 1, 1))
        {
          perror ("Initialisation failure in main");
          if (errno == EPERM)
            fprintf (stderr, "Command must be run as root\n");
          exit (EXIT_FAILURE);
        }
      errno = 0;
      if (ioperm (0xb2, 1, 1))
        {
          perror ("Initialisation failure in main");
          if (errno == EPERM)
            fprintf (stderr, "Command must be run as root\n");
          exit (EXIT_FAILURE);
        }
    }
  else
    {
      fprintf (stderr, "--------------????????  Unrecognised option:\n");
      typo ();
    }
  /* --info option  */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--info")))
      || (!(strcmp (argv[1], "-i")))))
    {
      err = get_dell_sig(0xfea3, &val1, &val2);
      if ((err == -EINVAL) || (err == -EAGAIN))
        fprintf (stderr, "Unable to read Dell sig1\n");
      else
        printf ("SIG1 eax = %c%c%c%c edx = %c%c%c%c\n", ((val1 >> 24) & 0xff),
                ((val1 >> 16) & 0xff), ((val1 >> 8) & 0xff), (val1 & 0xff),
                ((val2 >> 24) & 0xff), ((val2 >> 16) & 0xff),
                ((val2 >> 8) & 0xff), (val2 & 0xff));
      err = get_dell_sig(0xffa3, &val1, &val2);
      if ((err == -EINVAL) || (err == -EAGAIN))
        fprintf (stderr, "Unable to read Dell sig2\n");
      else
        printf ("SIG2 eax = %c%c%c%c edx = %c%c%c%c\n", ((val1 >> 24) & 0xff),
                ((val1 >> 16) & 0xff), ((val1 >> 8) & 0xff), (val1 & 0xff),
                ((val2 >> 24) & 0xff), ((val2 >> 16) & 0xff),
                ((val2 >> 8) & 0xff), (val2 & 0xff));
      val1 = get_bios_version ();
      if (val1 == -EINVAL) fprintf (stderr, "BIOS version not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading BIOS version\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("BIOS version = %c%c%c%c\n", ((val1 >> 24) & 0xff),
                ((val1 >> 16) & 0xff), ((val1 >> 8) & 0xff), (val1 & 0xff));
      val1 = get_cpu_temp ();
      if (val1 == -EINVAL) fprintf (stderr, "CPU temperature not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading CPU temperature\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("CPU temp = %d\n", val1);
      val1 = get_fan_status (0);
      if (val1 == -EINVAL) fprintf (stderr, "Fan 0 not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 0 status\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("Fan 0 status = %d\n", val1);
      val1 = get_fan_status (1);
      if (val1 == -EINVAL) fprintf (stderr, "Fan 1 not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 1 status\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("Fan 1 status = %d\n", val1);
      val1 = get_fan_speed (0);
      if (val1 == -EINVAL) fprintf (stderr, "Fan 0 RPM not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 0 RPM\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("Fan 0 RPM = %d\n", val1);
      val1 = get_fan_speed (1);
      if (val1 == -EINVAL) printf ("Fan 1 RPM not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 1 RPM\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("Fan 1 RPM = %d\n", val1);
      regs.eax = 0x0069;
      err = dell_smm (&regs);
      if (err == -EINVAL) printf ("Power status not found\n");
      if (err == -EAGAIN) fprintf (stderr, "Error reading power status\n");
      if (!err)
        {
          val1 = (regs.eax & 0xff);
          switch (val1)
            {
            case 5:
              printf ("Power supply = %d (AC mains)\n", val1);
              printf ("Battery charging current = %d\n", regs.ecx);
              break;
            case 3:
              printf ("Power supply = %d (Battery)\n", val1);
              break;
            case 0:
              printf ("Power supply = %d (AC, no battery detected)\n", val1);

            default:
              printf ("Power supply = %d\n", val1);
            }
          if (val1 != 0)
            {
              printf ("Battery nominal maximum capacity = %d\n", regs.edi);
              printf ("Battery current power remaining = %d\n", regs.edx);
            }
          printf ("Voltage = %dmV\n", regs.ebx);
        }
      val1 = get_docking_state ();
      if (val1 == -EINVAL) printf ("Dock status not found\n");
      if (val1 == -EAGAIN) fprintf (stderr, "Error reading dock status\n");
      if ((val1 != -EINVAL) && (val1 != -EAGAIN))
        printf ("Docking state = %d\n", val1);
      exit (EXIT_SUCCESS);
    }
  /* --daemon option arg parsing and validity checks */
  if ((argc >= 3) && ((!(strcmp (argv[1], "--daemon")))
      || (!(strcmp (argv[1], "-d")))))
    {
      dat.what_fan = strtol (argv[2], NULL, 0);
      if ((dat.what_fan >2) || (dat.what_fan < 0)) typo ();

      if (argc >= 4)
        {
          dat.poll = strtol (argv[3], NULL, 0);
          if ((dat.poll >30) || (dat.poll < 1)) typo ();
        }

      if (argc >= 5)
        {
          dat.t01 = strtol (argv[4], NULL, 0);
          if ((dat.t01 >= 127) || (dat.t01 < 0)) typo ();
        }
      if (argc >= 6)
        {
          dat.t02 = strtol (argv[5], NULL, 0);
          if ((dat.t02 >= 127) || (dat.t02 < 0)) typo ();
        }
      if (argc >= 7)
        {
          dat.h01 = strtol (argv[6], NULL, 0);
          if ((dat.h01 >= 127) || (dat.h01 < 0)) typo ();
        }
      if (argc >= 8)
        {
          dat.h02 = strtol (argv[7], NULL, 0);
          if ((dat.h02 >= 127) || (dat.h02 < 0)) typo ();
        }
      if (argc >= 9)
        {
          dat.t11 = strtol (argv[8], NULL, 0);
          if ((dat.t11 >= 127) || (dat.t11 < 0)) typo ();
        }
      if (argc >= 10)
        {
          dat.t12 = strtol (argv[9], NULL, 0);
          if ((dat.t12 >= 127) || (dat.t12 < 0)) typo ();
        }
      if (argc >= 11)
        {
          dat.h11 = strtol (argv[10], NULL, 0);
          if ((dat.h11 >= 127) || (dat.h11 < 0)) typo ();
        }
      if (argc >= 12)
        {
          dat.h12 = strtol (argv[11], NULL, 0);
          if ((dat.h12 >= 127) || (dat.h12 < 0)) typo ();
        }
      /* Args look valid, go and do daemon things */
      fan_daemon ();
    }
  /* --auto option, like --daemon but auto detects fan(s) */
  if ((argc >=2) && ((!(strcmp (argv[1], "--auto")))
      || (!(strcmp (argv[1], "-a")))))
    {
      for (tmp1 = 0; tmp1 < 1000; tmp1++)
        {
          val1 = get_fan_status (0);
          val2 = get_fan_status (1);
          if ((val1 != -EAGAIN) && (val2 != -EAGAIN)) break;
        }
      if (((val1 == -EAGAIN) || (val2 == -EAGAIN)) || ((val1 == -EINVAL)
          && (val2 == -EINVAL)))
        {
          fprintf (stderr, "fanxdell - WARNING - fans not detected\n");
          fprintf (stderr, "Try -d option for manual fan selection\n");
          fprintf (stderr, "BIOS must support ACPI. fanxdell terminating\n");
          exit (EXIT_FAILURE);
        } /* The above messages should go to syslog if printed at boot */
      if ((val1 != -EINVAL) && (val2 != -EINVAL))
        {
          dat.what_fan = 2;
        }
      else
        {
          if (val1 != -EINVAL) dat.what_fan = 0;
          if (val2 != -EINVAL) dat.what_fan = 1;
        }
      if (argc >= 3)
        {
          dat.poll = strtol (argv[2], NULL, 0);
          if ((dat.poll >30) || (dat.poll < 1)) typo ();
        }

      if (argc >= 4)
        {
          dat.t01 = strtol (argv[3], NULL, 0);
          if ((dat.t01 >= 127) || (dat.t01 < 0)) typo ();
        }
      if (argc >= 5)
        {
          dat.t02 = strtol (argv[4], NULL, 0);
          if ((dat.t02 >= 127) || (dat.t02 < 0)) typo ();
        }
      if (argc >= 6)
        {
          dat.h01 = strtol (argv[5], NULL, 0);
          if ((dat.h01 >= 127) || (dat.h01 < 0)) typo ();
        }
      if (argc >= 7)
        {
          dat.h02 = strtol (argv[6], NULL, 0);
          if ((dat.h02 >= 127) || (dat.h02 < 0)) typo ();
        }
      if (argc >= 8)
        {
          dat.t11 = strtol (argv[7], NULL, 0);
          if ((dat.t11 >= 127) || (dat.t11 < 0)) typo ();
        }
      if (argc >= 9)
        {
          dat.t12 = strtol (argv[8], NULL, 0);
          if ((dat.t12 >= 127) || (dat.t12 < 0)) typo ();
        }
      if (argc >= 10)
        {
          dat.h11 = strtol (argv[9], NULL, 0);
          if ((dat.h11 >= 127) || (dat.h11 < 0)) typo ();
        }
      if (argc >= 11)
        {
          dat.h12 = strtol (argv[10], NULL, 0);
          if ((dat.h12 >= 127) || (dat.h12 < 0)) typo ();
        }
      /* Args look valid, go and do daemon things */
      fan_daemon ();
    }
  /* --fan option to set fan speed(s) */
  if ((argc >= 3) && ((!(strcmp (argv[1], "--fan")))
      || (!(strcmp (argv[1], "-f")))))
    {
      /* set tmp1 */
      if (!(strcmp (argv[2], "-"))) tmp1 = -1;
      else tmp1 = strtol (argv[2], NULL, 0);
      /* set tmp2 (if there is one) */
      if (argc >= 4)
        {
          if (!(strcmp (argv[3], "-"))) tmp2 = -1;
          else tmp2 = strtol (argv[3], NULL, 0);
        }
      else
        {
          tmp2 = -1;
        }
      if ((tmp1 > 2) || (tmp1 < -1) || (tmp2 > 2) || (tmp2 < -1)) typo ();
      /* Set the fan(s) */
      if (tmp1 >= 0) val1 = set_fan(0, tmp1);
      if (val1 == -EINVAL) fprintf (stderr, "Unable to set fan 0\n");
      if (tmp2 >= 0) val2 = set_fan(1, tmp2);
      if (val2 == -EINVAL) fprintf (stderr, "Unable to set fan 1\n");
#ifdef DEBUG
      printf ("tmp1 = %d, tmp2 = %d\n", tmp1, tmp2);
      printf ("set_fan returned: val1 = %d, val2 = %d\n", val1, val2);
#endif
      exit (EXIT_SUCCESS);
    }
  /* --kill option */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--kill")))
      || (!(strcmp (argv[1], "-k"))))) kill_option ();
  /* End of options */
  fprintf (stderr, "Error parsing options\n");
  typo ();
  return (EXIT_FAILURE);
}
